from deepchem.utils.differentiation_utils import EditableModule
import torch
from abc import abstractmethod, abstractproperty
from typing import List


class BaseGrid(EditableModule):
    """
    BaseGrid is a class that regulates the integration points over the spatial
    dimensions.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.dft_utils import BaseGrid
    >>> class Grid(BaseGrid):
    ...     def __init__(self):
    ...         super(Grid, self).__init__()
    ...         self.ngrid = 10
    ...         self.ndim = 3
    ...         self.dvolume = torch.ones(self.ngrid, dtype=self.dtype, device=self.device)
    ...         self.rgrid = torch.ones((self.ngrid, self.ndim), dtype=self.dtype, device=self.device)
    ...     def get_dvolume(self):
    ...         return self.dvolume
    ...     def get_rgrid(self):
    ...         return self.rgrid
    >>> grid = Grid()
    >>> grid.get_dvolume()
    tensor([1., 1., 1., 1., 1., 1., 1., 1., 1., 1.])
    >>> grid.get_rgrid()
    tensor([[1., 1., 1.],
            [1., 1., 1.],
            [1., 1., 1.],
            [1., 1., 1.],
            [1., 1., 1.],
            [1., 1., 1.],
            [1., 1., 1.],
            [1., 1., 1.],
            [1., 1., 1.],
            [1., 1., 1.]])

    References
    ----------
    Kasim, Muhammad F., and Sam M. Vinko. "Learning the exchange-correlation
    functional from nature with fully differentiable density functional theory."
    Physical Review Letters 127.12 (2021): 126403.
    https://github.com/diffqc/dqc/blob/0fe821fc92cb3457fb14f6dff0c223641c514ddb/dqc/grid/base_grid.py

    """

    @abstractproperty
    def dtype(self) -> torch.dtype:
        """dtype of the grid points.

        Returns
        -------
        torch.dtype
            dtype of the grid points

        """
        pass

    @abstractproperty
    def device(self) -> torch.device:
        """device of the grid points

        Returns
        -------
        torch.device
            device of the grid points

        """
        pass

    @abstractproperty
    def coord_type(self) -> str:
        """type of the coordinate returned in get_rgrid. It can be 'cartesian'
        or 'spherical'.

        Returns
        -------
        str
            type of the coordinate returned in get_rgrid. It can be 'cartesian'
        or 'spherical'.

        """
        pass

    @abstractmethod
    def get_dvolume(self) -> torch.Tensor:
        """Obtain the torch.tensor containing the dV elements for the integration.

        Returns
        -------
        torch.tensor (*BG, ngrid)
            The dV elements for the integration. *BG is the length of the BaseGrid.

        """
        pass

    @abstractmethod
    def get_rgrid(self) -> torch.Tensor:
        """
        Returns the grid points position in the specified coordinate in
        self.coord_type.

        Returns
        -------
        torch.tensor (*BG, ngrid, ndim)
            The grid points position. *BG is the length of the BaseGrid.

        """
        pass

    @abstractmethod
    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """
        Return a list with the parameter names corresponding to the given method
        (methodname)

        Returns
        -------
        List[str]
            List of parameter names of methodname

        """
        pass



"""Utilities for handling PDBQT files."""

from typing import Dict, List, Optional, Set, Tuple
from deepchem.utils.typing import RDKitMol  # type: ignore


def pdbqt_to_pdb(filename: Optional[str] = None,
                 pdbqt_data: Optional[List[str]] = None) -> str:
    """Extracts the PDB part of a pdbqt file as a string.

    Either `filename` or `pdbqt_data` must be provided. This function
    strips PDBQT charge information from the provided input.

    Parameters
    ----------
    filename: str, optional  (default None)
        Filename of PDBQT file
    pdbqt_data: List[str], optional (default None)
        Raw list of lines containing data from PDBQT file.

    Returns
    -------
    pdb_block: str
        String containing the PDB portion of pdbqt file.
    """
    if filename is not None and pdbqt_data is not None:
        raise ValueError("Only one of filename or pdbqt_data can be provided")
    elif filename is None and pdbqt_data is None:
        raise ValueError("Either filename or pdbqt_data must be provided")
    elif filename is not None:
        pdbqt_data = open(filename).readlines()

    pdb_block = ""
    # FIXME: Item "None" of "Optional[List[str]]" has no attribute "__iter__" (not iterable)
    for line in pdbqt_data:  # type: ignore
        pdb_block += "%s\n" % line[:66]
    return pdb_block


def convert_protein_to_pdbqt(mol: RDKitMol, outfile: str) -> None:
    """Convert a protein PDB file into a pdbqt file.

    Writes the extra PDBQT terms directly to `outfile`.

    Parameters
    ----------
    mol: rdkit.Chem.rdchem.Mol
        Protein molecule
    outfile: str
        filename which already has a valid pdb representation of mol
    """
    lines = [x.strip() for x in open(outfile).readlines()]
    out_lines = []
    for line in lines:
        if "ROOT" in line or "ENDROOT" in line or "TORSDOF" in line:
            out_lines.append("%s\n" % line)
            continue
        if not line.startswith("ATOM"):
            continue
        line = line[:66]
        atom_index = int(line[6:11])
        atom = mol.GetAtoms()[atom_index - 1]
        line = "%s    +0.000 %s\n" % (line, atom.GetSymbol().ljust(2))
        out_lines.append(line)
    with open(outfile, 'w') as fout:
        for line in out_lines:
            fout.write(line)


def _mol_to_graph(mol: RDKitMol):
    """Convert RDKit Mol to NetworkX graph

    Convert mol into a graph representation atoms are nodes, and bonds
    are vertices stored as graph

    Parameters
    ----------
    mol: rdkit.Chem.rdchem.Mol
        The molecule to convert into a graph.

    Returns
    -------
    graph: networkx.Graph
        Contains atoms indices as nodes, edges as bonds.

    Notes
    -----
    This function requires NetworkX to be installed.
    """
    try:
        import networkx as nx  # type: ignore
    except ModuleNotFoundError:
        raise ImportError("This function requires NetworkX to be installed.")

    G = nx.Graph()
    num_atoms = mol.GetNumAtoms()
    G.add_nodes_from(range(num_atoms))
    for i in range(mol.GetNumBonds()):
        from_idx = mol.GetBonds()[i].GetBeginAtomIdx()
        to_idx = mol.GetBonds()[i].GetEndAtomIdx()
        G.add_edge(from_idx, to_idx)
    return G


def _get_rotatable_bonds(mol: RDKitMol) -> List[Tuple[int, int]]:
    """
    https://github.com/rdkit/rdkit/blob/f4529c910e546af590c56eba01f96e9015c269a6/Code/GraphMol/Descriptors/Lipinski.cpp#L107

    Taken from rdkit source to find which bonds are rotatable store
    rotatable bonds in (from_atom, to_atom)

    Parameters
    ----------
    mol: rdkit.Chem.rdchem.Mol
        Ligand molecule

    Returns
    -------
    rotatable_bonds: List[List[int, int]]
        List of rotatable bonds in molecule

    Notes
    -----
    This function requires RDKit to be installed.
    """
    try:
        from rdkit import Chem  # type: ignore
        from rdkit.Chem import rdmolops  # type: ignore
    except ModuleNotFoundError:
        raise ImportError("This function requires RDKit to be installed.")

    pattern = Chem.MolFromSmarts(
        "[!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])("
        "[CH3])[CH3])&!$([CD3](=[N,O,S])-!@[#7,O,S!D1])&!$([#7,O,S!D1]-!@[CD3]="
        "[N,O,S])&!$([CD3](=[N+])-!@[#7!D1])&!$([#7!D1]-!@[CD3]=[N+])]-!@[!$(*#"
        "*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])([CH3])"
        "[CH3])]")
    rdmolops.FastFindRings(mol)
    rotatable_bonds = mol.GetSubstructMatches(pattern)
    return rotatable_bonds


def convert_mol_to_pdbqt(mol: RDKitMol, outfile: str) -> None:
    """Writes the provided ligand molecule to specified file in pdbqt format.

    Creates a torsion tree and write to pdbqt file. The torsion tree
    represents rotatable bonds in the molecule.

    Parameters
    ----------
    mol: rdkit.Chem.rdchem.Mol
        The molecule whose value is stored in pdb format in outfile
    outfile: str
        Filename for a valid pdb file with the extention .pdbqt

    Notes
    -----
    This function requires NetworkX to be installed.
    """
    try:
        import networkx as nx
    except ModuleNotFoundError:
        raise ImportError("This function requires NetworkX to be installed.")

    # Walk through the original file and extract ATOM/HETATM lines and
    # add PDBQT charge annotations.
    pdb_map = _create_pdb_map(outfile)
    graph = _mol_to_graph(mol)
    rotatable_bonds = _get_rotatable_bonds(mol)

    # Remove rotatable bonds from this molecule
    for bond in rotatable_bonds:
        graph.remove_edge(bond[0], bond[1])
    # Get the connected components now that the rotatable bonds have
    # been removed.
    components = [x for x in nx.connected_components(graph)]
    comp_map = _create_component_map(mol, components)

    used_partitions = set()
    lines = []
    # The root is the largest connected component.
    root = max(enumerate(components), key=lambda x: len(x[1]))[0]
    # Write the root component
    lines.append("ROOT\n")
    for atom in components[root]:
        lines.append(pdb_map[atom])
    lines.append("ENDROOT\n")
    # We've looked at the root, so take note of that
    used_partitions.add(root)
    for bond in rotatable_bonds:
        valid, next_partition = _valid_bond(used_partitions, bond, root,
                                            comp_map)
        if not valid:
            continue
        _dfs(used_partitions, next_partition, bond, components, rotatable_bonds,
             lines, pdb_map, comp_map)
    lines.append("TORSDOF %s" % len(rotatable_bonds))
    with open(outfile, 'w') as fout:
        for line in lines:
            fout.write(line)


def _create_pdb_map(outfile: str) -> Dict[int, str]:
    """Create a mapping from atom numbers to lines to write to pdbqt

    This is a map from rdkit atom number to its line in the pdb
    file. We also add the two additional columns required for
    pdbqt (charge, symbol).

    Note rdkit atoms are 0 indexed and pdb files are 1 indexed

    Parameters
    ----------
    outfile: str
        filename which already has a valid pdb representation of mol

    Returns
    -------
    pdb_map: Dict[int, str]
        Maps rdkit atom numbers to lines to be written to PDBQT file.
    """
    lines = [x.strip() for x in open(outfile).readlines()]
    lines = list(
        filter(lambda x: x.startswith("HETATM") or x.startswith("ATOM"), lines))
    lines = [x[:66] for x in lines]
    pdb_map = {}
    for line in lines:
        my_values = line.split()
        atom_number = int(my_values[1])
        atom_symbol = my_values[2].capitalize()
        atom_symbol = ''.join([i for i in atom_symbol if not i.isdigit()])
        line = line.replace("HETATM", "ATOM  ")
        line = "%s    +0.000 %s\n" % (line, atom_symbol.ljust(2))
        pdb_map[atom_number - 1] = line
    return pdb_map


def _create_component_map(mol: RDKitMol,
                          components: List[List[int]]) -> Dict[int, int]:
    """Creates a map from atom ids to disconnected component id

    For each atom in `mol`, maps it to the id of the component in the
    molecule. The intent is that this is used on a molecule whose
    rotatable bonds have been removed. `components` is a list of the
    connected components after this surgery.

    Parameters
    ----------
    mol: rdkit.Chem.rdchem.Mol
        The molecule to find disconnected components in
    components: List[List[int]]
        List of connected components

    Returns
    -------
    comp_map: Dict[int, int]
        Maps atom ids to component ides
    """
    comp_map = {}
    for i in range(mol.GetNumAtoms()):
        for j in range(len(components)):
            if i in components[j]:
                comp_map[i] = j
                break
    return comp_map


def _dfs(used_partitions: Set[int], current_partition: int,
         bond: Tuple[int, int], components: List[List[int]],
         rotatable_bonds: List[Tuple[int, int]], lines: List[str],
         pdb_map: Dict[int, str], comp_map: Dict[int, int]) -> List[str]:
    """
    This function does a depth first search through the torsion tree

    Parameters
    ----------
    used_partions: Set[int]
        Partitions which have already been used
    current_partition: int
        The current partition to expand
    bond: Tuple[int, int]
        the bond which goes from the previous partition into this partition
    components: List[List[int]]
        List of connected components
    rotatable_bonds: List[Tuple[int, int]]
        List of rotatable bonds. This tuple is (from_atom, to_atom).
    lines: List[str]
        List of lines to write
    pdb_map: Dict[int, str]
        Maps atom numbers to PDBQT lines to write
    comp_map: Dict[int, int]
        Maps atom numbers to component numbers

    Returns
    -------
    lines: List[str]
        List of lines to write. This has more appended lines.
    """
    if comp_map[bond[1]] != current_partition:
        bond = (bond[1], bond[0])
    used_partitions.add(comp_map[bond[0]])
    used_partitions.add(comp_map[bond[1]])
    lines.append("BRANCH %4s %4s\n" % (bond[0] + 1, bond[1] + 1))
    for atom in components[current_partition]:
        lines.append(pdb_map[atom])
    for b in rotatable_bonds:
        valid, next_partition = \
          _valid_bond(used_partitions, b, current_partition, comp_map)
        if not valid:
            continue
        lines = _dfs(used_partitions, next_partition, b, components,
                     rotatable_bonds, lines, pdb_map, comp_map)
    lines.append("ENDBRANCH %4s %4s\n" % (bond[0] + 1, bond[1] + 1))
    return lines


def _valid_bond(used_partitions: Set[int], bond: Tuple[int, int],
                current_partition: int,
                comp_map: Dict[int, int]) -> Tuple[bool, int]:
    """Helper method to find next partition to explore.

    Used to check if a bond goes from the current partition into a
    partition that is not yet explored

    Parameters
    ----------
    used_partions: Set[int]
        Partitions which have already been used
    bond: Tuple[int, int]
        The bond to check if it goes to an unexplored partition.
        This tuple is (from_atom, to_atom).
    current_partition: int
        The current partition of the DFS
    comp_map: Dict[int, int]
        Maps atom ids to component ids

    Returns
    -------
    is_valid: bool
        Whether to exist the next partition or not
    next_partition: int
        The next partition to explore
    """
    part1 = comp_map[bond[0]]
    part2 = comp_map[bond[1]]
    if part1 != current_partition and part2 != current_partition:
        return False, 0
    if part1 == current_partition:
        next_partition = part2
    else:
        next_partition = part1
    return next_partition not in used_partitions, next_partition



from deepchem.models.torch_models.modular import TorchModel
from deepchem.models.losses import BinaryCrossEntropy
import torch
import torch.nn as nn
import torch.nn.functional as F


class UNet(nn.Module):
    """
    UNet model for image segmentation.

    UNet is a convolutional neural network architecture for fast and precise segmentation of images
    based on the works of Ronneberger et al. [1]. The architecture consists of an encoder, a bottleneck,
    and a decoder. The encoder downsamples the input image to capture the context of the image. The
    bottleneck captures the most important features of the image. The decoder upsamples the image to
    generate the segmentation mask. The encoder and decoder are connected by skip connections to preserve
    spatial information.

    Examples
    --------
    Importing necessary modules

    >>> import numpy as np
    >>> import deepchem as dc
    >>> from deepchem.models.torch_models import UNet

    Creating a random dataset of 5 32x32 pixel RGB input images and 5 32x32 pixel grey scale output images

    >>> x = np.random.randn(5, 3, 32, 32).astype(np.float32)
    >>> y = np.random.rand(5, 1, 32, 32).astype(np.float32)
    >>> dataset = dc.data.NumpyDataset(x, y)

    We will create a UNet model with 3 input channels and 1 output channel. We will then fit the model on the dataset for 5 epochs and predict the output images.

    >>> model = UNetModel(in_channels=3, out_channels=1)
    >>> model.fit(dataset, nb_epoch=5)
    >>> predictions = model.predict(dataset)

    Notes
    -----
    1. This implementation of the UNet model makes some changes to the padding of the inputs to the convolutional layers.
    The padding is set to 'same' to ensure that the output size of the convolutional layers is the same as the input size.
    This is done to preserve the spatial information of the input image and to keep the output size of the encoder and decoder the same.

    2. The input image size must be divisible by 2^4 = 16 to ensure that the output size of the encoder and decoder is the same.

    References
    ----------
    .. [1] Ronneberger, O., Fischer, P., & Brox, T. (2015, May 18). U-NET: Convolutional Networks for Biomedical Image Segmentation. arXiv.org. https://arxiv.org/abs/1505.04597
    """

    def __init__(self, in_channels: int = 3, out_channels: int = 1):
        """
        Parameters
        ----------
        in_channels: int (default 3)
            Number of input channels.
        out_channels: int (default 1)
            Number of output channels.
        """
        super(UNet, self).__init__()

        # Encoder
        self.encoder1 = self.conv_block(in_channels, 64)
        self.encoder2 = self.conv_block(64, 128)
        self.encoder3 = self.conv_block(128, 256)
        self.encoder4 = self.conv_block(256, 512)

        # Bottleneck
        self.bottleneck = self.conv_block(512, 1024)

        # Decoder
        self.decoder4 = self.conv_block(1024 + 512, 512)
        self.decoder3 = self.conv_block(512 + 256, 256)
        self.decoder2 = self.conv_block(256 + 128, 128)
        self.decoder1 = self.conv_block(128 + 64, 64)

        # Maxpooling
        self.maxpool = nn.MaxPool2d(kernel_size=2, stride=2)

        # Upsampling
        self.upsample = nn.Upsample(scale_factor=2,
                                    mode='bilinear',
                                    align_corners=True)

        # Output
        self.output = nn.Conv2d(64, out_channels, kernel_size=1)

    def conv_block(self, in_channels: int, out_channels: int):
        """
        Parameters
        ----------
        in_channels: int
            Number of input channels.
        out_channels: int
            Number of output channels.
        """
        return nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, padding='same'),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, kernel_size=3,
                      padding='same'), nn.ReLU(inplace=True))

    def forward(self, x: torch.Tensor):
        """
        Parameters
        ----------
        x: Tensor
            Input tensor.

        Returns
        -------
        x: Tensor
            Output tensor.
        """

        # Encoder
        x1 = self.encoder1(x)
        x = self.maxpool(x1)
        x2 = self.encoder2(x)
        x = self.maxpool(x2)
        x3 = self.encoder3(x)
        x = self.maxpool(x3)
        x4 = self.encoder4(x)
        x = self.maxpool(x4)

        # Bottleneck
        x = self.bottleneck(x)

        # Decoder
        x = self.upsample(x)
        x = self.decoder4(torch.cat([x, x4], 1))
        x = self.upsample(x)
        x = self.decoder3(torch.cat([x, x3], 1))
        x = self.upsample(x)
        x = self.decoder2(torch.cat([x, x2], 1))
        x = self.upsample(x)
        x = self.decoder1(torch.cat([x, x1], 1))

        # Output
        x = self.output(x)
        x = F.sigmoid(x)
        return x


class UNetModel(TorchModel):
    """
    UNet model for image segmentation.

    UNet is a convolutional neural network architecture for fast and precise segmentation of images
    based on the works of Ronneberger et al. [1]. The architecture consists of an encoder, a bottleneck,
    and a decoder. The encoder downsamples the input image to capture the context of the image. The
    bottleneck captures the most important features of the image. The decoder upsamples the image to
    generate the segmentation mask. The encoder and decoder are connected by skip connections to preserve
    spatial information.

    Examples
    --------
    Importing necessary modules

    >>> import numpy as np
    >>> import deepchem as dc
    >>> from deepchem.models.torch_models import UNet

    Creating a random dataset of 5 32x32 pixel RGB input images and 5 32x32 pixel grey scale output images

    >>> x = np.random.randn(5, 3, 32, 32).astype(np.float32)
    >>> y = np.random.rand(5, 1, 32, 32).astype(np.float32)
    >>> dataset = dc.data.NumpyDataset(x, y)

    We will create a UNet model with 3 input channels and 1 output channel. We will then fit the model on the dataset for 5 epochs and predict the output images.

    >>> model = UNetModel(in_channels=3, out_channels=1)
    >>> model.fit(dataset, nb_epoch=5)
    >>> predictions = model.predict(dataset)

    Notes
    -----
    1. This implementation of the UNet model makes some changes to the padding of the inputs to the convolutional layers.
    The padding is set to 'same' to ensure that the output size of the convolutional layers is the same as the input size.
    This is done to preserve the spatial information of the input image and to keep the output size of the encoder and decoder the same.

    2. The input image size must be divisible by 2^4 = 16 to ensure that the output size of the encoder and decoder is the same.

    References
    ----------
    .. [1] Ronneberger, O., Fischer, P., & Brox, T. (2015, May 18). U-NET: Convolutional Networks for Biomedical Image Segmentation. arXiv.org. https://arxiv.org/abs/1505.04597
    """

    def __init__(self, in_channels: int = 3, out_channels: int = 1, **kwargs):
        """
        Parameters
        ----------
        input_channels: int (default 3)
            Number of input channels.
        output_channels: int (default 1)
            Number of output channels.
        """
        if in_channels <= 0:
            raise ValueError("input_channels must be greater than 0")

        if out_channels <= 0:
            raise ValueError("output_channels must be greater than 0")

        model = UNet(in_channels=in_channels, out_channels=out_channels)

        if 'loss' not in kwargs:
            kwargs['loss'] = BinaryCrossEntropy()

        super(UNetModel, self).__init__(model, **kwargs)



import deepchem as dc
import numpy as np
import pytest
try:
    import torch
    has_torch = True
except ModuleNotFoundError:
    has_torch = False


@pytest.mark.torch
def test_unsorted_segment_sum():

    segment_ids = torch.Tensor([0, 1, 0]).to(torch.int64)
    data = torch.Tensor([[1, 2, 3, 4], [5, 6, 7, 8], [4, 3, 2, 1]])
    num_segments = 2

    if len(segment_ids.shape) != 1:
        raise AssertionError("segment_ids have be a 1-D tensor")

    if data.shape[0] != segment_ids.shape[0]:
        raise AssertionError(
            "segment_ids should be the same size as dimension 0 of input.")

    result = dc.utils.pytorch_utils.unsorted_segment_sum(
        data=data, segment_ids=segment_ids, num_segments=num_segments)

    assert np.allclose(
        np.array(result),
        np.load("deepchem/utils/test/assets/result_segment_sum.npy"),
        atol=1e-04)


@pytest.mark.torch
def test_segment_sum():

    data = torch.Tensor([[1, 2, 3, 4], [4, 3, 2, 1], [5, 6, 7, 8]])
    segment_ids = torch.Tensor([0, 0, 1]).to(torch.int64)

    if len(segment_ids.shape) != 1:
        raise AssertionError("segment_ids have be a 1-D tensor")

    if data.shape[0] != segment_ids.shape[0]:
        raise AssertionError(
            "segment_ids should be the same size as dimension 0 of input.")

    result = dc.utils.pytorch_utils.segment_sum(data=data,
                                                segment_ids=segment_ids)

    assert np.allclose(
        np.array(result),
        np.load("deepchem/utils/test/assets/result_segment_sum.npy"),
        atol=1e-04)


@pytest.mark.torch
def test_chunkify():
    """Test chunkify utility."""
    data = torch.arange(10)
    chunked_list = list(dc.utils.pytorch_utils.chunkify(data, 0, 3))
    assert len(chunked_list) == 4
    assert chunked_list[0][0].tolist() == [0, 1, 2]
    assert chunked_list[3][0].tolist() == [9]
    assert chunked_list[0][1] == 0
    assert chunked_list[3][1] == 9


@pytest.mark.torch
def test_get_memory():
    """Test get_memory utility."""
    data = torch.rand(100, 100, dtype=torch.float64)
    assert dc.utils.pytorch_utils.get_memory(data) == 100 * 100 * 8


@pytest.mark.torch
def test_gaussian_integral():
    """Test the gaussian integral utility."""
    assert dc.utils.pytorch_utils.gaussian_integral(5, 1.0) == 1.0


@pytest.mark.torch
def test_TensorNonTensorSeparator():
    """Test the TensorNonTensorSeparator utility."""
    a = torch.tensor([1., 2, 3])
    b = 4.
    c = torch.tensor([5., 6, 7], requires_grad=True)
    params = [a, b, c]
    separator = dc.utils.pytorch_utils.TensorNonTensorSeparator(params)
    tensor_params = separator.get_tensor_params()
    assert torch.allclose(tensor_params[0],
                          torch.tensor([5., 6., 7.], requires_grad=True))


@pytest.mark.torch
def test_tallqr():
    V = torch.randn(3, 2)
    Q, R = dc.utils.pytorch_utils.tallqr(V)
    assert Q.shape == torch.Size([3, 2])
    assert R.shape == torch.Size([2, 2])
    assert torch.allclose(Q @ R, V)


@pytest.mark.torch
def test_to_fortran_order():
    V = torch.randn(3, 2)
    if V.is_contiguous() is False:
        assert False
    V = dc.utils.pytorch_utils.to_fortran_order(V)
    if V.is_contiguous() is True:
        assert False
    assert V.shape == torch.Size([3, 2])


@pytest.mark.torch
def test_get_np_dtype():
    """Test the get_np_dtype utility."""
    assert dc.utils.pytorch_utils.get_np_dtype(torch.float32) == np.float32
    assert dc.utils.pytorch_utils.get_np_dtype(torch.float64) == np.float64



import os
import unittest
import numpy as np
from deepchem.utils import rdkit_utils
from deepchem.utils.fragment_utils import get_contact_atom_indices
from deepchem.utils.fragment_utils import merge_molecular_fragments
from deepchem.utils.fragment_utils import get_partial_charge
from deepchem.utils.fragment_utils import strip_hydrogens
from deepchem.utils.fragment_utils import MolecularFragment
from deepchem.utils.fragment_utils import AtomShim


class TestFragmentUtil(unittest.TestCase):

    def setUp(self):
        # TODO test more formats for ligand
        current_dir = os.path.dirname(os.path.realpath(__file__))
        self.protein_file = os.path.join(
            current_dir, '../../feat/tests/data/3ws9_protein_fixer_rdkit.pdb')
        self.ligand_file = os.path.join(
            current_dir, '../../feat/tests/data/3ws9_ligand.sdf')

    def test_get_contact_atom_indices(self):
        complexes = rdkit_utils.load_complex(
            [self.protein_file, self.ligand_file])
        contact_indices = get_contact_atom_indices(complexes)
        assert len(contact_indices) == 2

    def test_create_molecular_fragment(self):
        mol_xyz, mol_rdk = rdkit_utils.load_molecule(self.ligand_file)
        fragment = MolecularFragment(mol_rdk.GetAtoms(), mol_xyz)
        assert len(mol_rdk.GetAtoms()) == len(fragment.GetAtoms())
        assert (fragment.GetCoords() == mol_xyz).all()

    def test_strip_hydrogens(self):
        mol_xyz, mol_rdk = rdkit_utils.load_molecule(self.ligand_file)
        _ = MolecularFragment(mol_rdk.GetAtoms(), mol_xyz)

        # Test on RDKit
        _ = strip_hydrogens(mol_xyz, mol_rdk)

    def test_merge_molecular_fragments(self):
        mol_xyz, mol_rdk = rdkit_utils.load_molecule(self.ligand_file)
        fragment1 = MolecularFragment(mol_rdk.GetAtoms(), mol_xyz)
        fragment2 = MolecularFragment(mol_rdk.GetAtoms(), mol_xyz)
        joint = merge_molecular_fragments([fragment1, fragment2])
        assert len(mol_rdk.GetAtoms()) * 2 == len(joint.GetAtoms())

    def test_get_partial_charge(self):
        from rdkit import Chem
        mol = Chem.MolFromSmiles("CC")
        atom = mol.GetAtoms()[0]
        partial_charge = get_partial_charge(atom)
        assert partial_charge == 0

    def test_atom_shim(self):
        atomic_num = 5
        partial_charge = 1
        atom_coords = np.array([0., 1., 2.])
        shim = AtomShim(atomic_num, partial_charge, atom_coords)
        assert shim.GetAtomicNum() == atomic_num
        assert shim.GetPartialCharge() == partial_charge
        assert (shim.GetCoords() == atom_coords).all()



"""
Density Functional Theory utilities.
"""
# flake8: noqa
import logging

logger_ = logging.getLogger(__name__)

from deepchem.utils.dft_utils.config import config

try:
    from deepchem.utils.dft_utils.hamilton.intor.lattice import Lattice

    from deepchem.utils.dft_utils.data.datastruct import ZType
    from deepchem.utils.dft_utils.data.datastruct import AtomPosType
    from deepchem.utils.dft_utils.data.datastruct import AtomZsType
    from deepchem.utils.dft_utils.data.datastruct import SpinParam
    from deepchem.utils.dft_utils.data.datastruct import ValGrad
    from deepchem.utils.dft_utils.data.datastruct import CGTOBasis
    from deepchem.utils.dft_utils.data.datastruct import AtomCGTOBasis
    from deepchem.utils.dft_utils.data.datastruct import BasisInpType

    from deepchem.utils.dft_utils.hamilton.orbparams import BaseOrbParams
    from deepchem.utils.dft_utils.hamilton.orbparams import QROrbParams
    from deepchem.utils.dft_utils.hamilton.orbparams import MatExpOrbParams

    from deepchem.utils.dft_utils.api.parser import parse_moldesc

    from deepchem.utils.dft_utils.grid.base_grid import BaseGrid

    from deepchem.utils.dft_utils.grid.radial_grid import RadialGrid
    from deepchem.utils.dft_utils.grid.radial_grid import get_xw_integration
    from deepchem.utils.dft_utils.grid.radial_grid import SlicedRadialGrid
    from deepchem.utils.dft_utils.grid.radial_grid import BaseGridTransform
    from deepchem.utils.dft_utils.grid.radial_grid import DE2Transformation
    from deepchem.utils.dft_utils.grid.radial_grid import LogM3Transformation
    from deepchem.utils.dft_utils.grid.radial_grid import TreutlerM4Transformation
    from deepchem.utils.dft_utils.grid.radial_grid import get_grid_transform

    from deepchem.utils.dft_utils.xc.base_xc import BaseXC
    from deepchem.utils.dft_utils.xc.base_xc import AddBaseXC
    from deepchem.utils.dft_utils.xc.base_xc import MulBaseXC
    from deepchem.utils.dft_utils.xc.libxc_wrapper import CalcLDALibXCPol
    from deepchem.utils.dft_utils.xc.libxc_wrapper import CalcLDALibXCUnpol
    from deepchem.utils.dft_utils.xc.libxc_wrapper import CalcGGALibXCUnpol
    from deepchem.utils.dft_utils.xc.libxc_wrapper import CalcGGALibXCPol
    from deepchem.utils.dft_utils.xc.libxc_wrapper import CalcMGGALibXCUnpol
    from deepchem.utils.dft_utils.xc.libxc_wrapper import CalcMGGALibXCPol
    from deepchem.utils.dft_utils.xc.libxc import LibXCLDA
    from deepchem.utils.dft_utils.xc.libxc import LibXCGGA
    from deepchem.utils.dft_utils.xc.libxc import LibXCMGGA

    from deepchem.utils.dft_utils.api.getxc import get_libxc
    from deepchem.utils.dft_utils.api.getxc import get_xc

    from deepchem.utils.dft_utils.df.base_df import BaseDF

    from deepchem.utils.dft_utils.hamilton.base_hamilton import BaseHamilton

    from deepchem.utils.dft_utils.system.base_system import BaseSystem

    from deepchem.utils.dft_utils.qccalc.base_qccalc import BaseQCCalc
    from deepchem.utils.dft_utils.qccalc.scf_qccalc import SCF_QCCalc
    from deepchem.utils.dft_utils.qccalc.scf_qccalc import BaseSCFEngine
    from deepchem.utils.dft_utils.qccalc.hf import HF
    from deepchem.utils.dft_utils.qccalc.hf import HFEngine
except ModuleNotFoundError as e:
    logger_.warning(
        f'Skipped loading some Pytorch utilities, missing a dependency. {e}')
